/**
  Tiny OpenGL/GLUT demo of GPU texture computation.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-01-21 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "GL/glui.h" /* user interface widgets: http://glui.sourceforge.net/ */

#include "soil/SOIL.h" /* Simple OpenGL Image Library, www.lonesock.net/soil.html (plus Dr. Lawlor's modifications) */
#include "soil/SOIL.c" /* just slap in implementation files here, for easier linking */
#include "soil/stb_image_aug.c"

#include "ogl/glsl.h"
#include "ogl/glsl.cpp"

#include "ogl/minicam.h" /* ortho_frame, camera motion, key_down & events */

#include <vector> /* std::vector container */

#ifndef M_PI /* why doesn't windows math.h have this? */
#define M_PI 3.14159265358979
#endif

/* These variables are changed directly by the user interface code below */

/* Enable for anisotropic filtering */
int texaniso=1;

/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; 

/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_REPEAT;

/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; 
int magfilter_mode=GL_LINEAR; 

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_texture(int w,int h,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	char *mem=new char[w*h*4]; memset(mem,0x80,w*h*4);
	glTexImage2D(t,0,format, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,GL_LINEAR); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,GL_LINEAR); /* no mipmaps */
	GLenum texWrap=GL_CLAMP_TO_EDGE; /* stop changing at edge */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	glGenerateMipmapEXT(t); /* make mipmaps */
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint tex)
{
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}

void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.01, display_dt=0.0;
	static float sim_time=0.0;
	float big=300; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	
/* Do offscreen rendering work */
	/* load an image file directly as a new OpenGL texture */
	GLenum texture_format=GL_RGBA8; // or try GL_RGBA_FLOAT16_ATI;
	static GLuint srcTex = SOIL_load_OGL_texture
	(
		"image.jpg",
		SOIL_LOAD_AUTO,
		SOIL_CREATE_NEW_ID,
		SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y,
		texture_format
	);
	int w=512, h=512;
	static GLuint destTex = make_texture(w,h,texture_format);
	static GLuint fb = make_fbo(destTex);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	
/* Grab the highest mipmap level */
	int level=0;
	int size=w;
	while (size!=1) { level++; size/=2; }
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, srcTex, level);
	float pixel[4];
	glReadPixels(0,0,1,1,GL_RGBA,GL_FLOAT,pixel);
	//printf("Average height %.4f, velocity %.4f\n",pixel[0],pixel[1]);
	float stimulus = 0.5-pixel[0]; /* force the average height to be 0.5 */
	
/* Render next step of simulation */
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, destTex, 0);
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	/* From 0..1 texture coords to -1..+1 clip coords */
	glTranslatef(-1.0,-1.0,-1.0);
	glScalef(2.0,2.0,2.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glBindTexture(GL_TEXTURE_2D,srcTex);
	glEnable(GL_BLEND);

	/* Actual physics happens right on the graphics card */
	static GLhandleARB prog = makeProgramObject(
	"varying vec2 texCoords;\n"
	"void main(void) /* vertex shader */ {\n"
	"	texCoords = vec2(gl_MultiTexCoord0);\n"
	"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
	"}\n"
	,
	"varying vec2 texCoords;\n"
	"uniform sampler2D srcTex;\n"
	"uniform float stimulus;\n"
	"const int dirs=4; /* MUST match C++ copy below! */\n"
	"uniform vec2 sampleDirs[dirs]; /* texture coordinate shifts in each direction */\n"
	"void main(void) /* fragment shader */ {\n"
	"	vec4 cen = texture2D(srcTex,texCoords); /* center pixel */\n"
	"	vec4 nbor[dirs]; /* neighboring pixels */\n"
	"	for (int d=0;d<dirs;d++) {\n"
	"		nbor[d]=texture2D(srcTex,texCoords+sampleDirs[d]);\n"
	"	}\n"
	"	vec4 fxx = (nbor[2]-2.0*cen+nbor[0]);\n"
	"	vec4 fyy = (nbor[3]-2.0*cen+nbor[1]);\n"
	"	float dt=0.2, scl=5.0; /* timestep, velocity scale onscreen*/\n"
	"	float vel = (cen.g-cen.b)*(1.0/scl); \n"
	"	float acc = fxx+fyy; \n"
	"	vel += dt*acc;\n"
	"	cen.r += stimulus + dt*vel + dt*dt*acc*0.5;\n"
	"	gl_FragColor = clamp(vec4(cen.r,vel*scl,-vel*scl,1.0), 0.0,1.0);\n"
	"}\n"
	);
	glUseProgramObjectARB(prog);
	glUniform1iARB(glGetUniformLocationARB(prog,"srcTex"),0); /* texture unit 0 */
	glUniform1fARB(glGetUniformLocationARB(prog,"stimulus"),stimulus);

	enum {dirs=4}; /* MUST match GLSL version above! */
	float sampleDirs[2*dirs];
	for (int d=0;d<dirs;d++) {
		float r=1.0/w; /* filter size, in texture coordinates */
		float angle=d*(2.0*M_PI/dirs); /* filter orientation, radians */
		sampleDirs[2*d+0]=r*cos(angle);
		sampleDirs[2*d+1]=r*sin(angle);
		//printf("sampleDirs[%d] = %f, %f\n",d,sampleDirs[2*d+0],sampleDirs[2*d+1]);
	}
	glUniform2fvARB(glGetUniformLocationARB(prog,"sampleDirs"),dirs,sampleDirs);
	
	/* run this program over the whole rectangle of the texture */
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	glTexCoord2f(0,0); glVertex2fv(vec3(0,0,0));
	glTexCoord2f(1,0); glVertex2fv(vec3(1,0,0));
	glTexCoord2f(1,1); glVertex2fv(vec3(1,1,0));
	glTexCoord2f(0,1); glVertex2fv(vec3(0,1,0));
	glEnd();
	
	glUseProgramObjectARB(0);
	
	if (key_down[(int)' ']) { /* interact with the simulation */
		float x=mouse_x/(float)glutGet(GLUT_WINDOW_WIDTH);
		float y=1.0-mouse_y/(float)glutGet(GLUT_WINDOW_HEIGHT);
		glTranslatef(x,y,0);
		glDisable(GL_TEXTURE_2D);
		glColor4f(1,0.0,0.0,1.0);
		glutSolidTeapot(0.1);
	}
	

	if (1) { /* limit framerate */
#ifdef __unix__
	usleep(10*1000); /* sleep for 10ms, to cut framerate to 100hz */
#else /* windows? */
	Sleep(10);
#endif	
	}
	glDisable(GL_TEXTURE_2D);
	
	std::swap(srcTex,destTex); /* ping-pong the buffers */
	
	oglCheck("After offscreen work");
	
/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	oglCameraLookAt(0.5);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
// Draw in our ground plane texture
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,srcTex);
	glGenerateMipmapEXT(t); /* make mipmaps before rendering */
	glEnable(t);
	
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	glTexParameteri(t, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(t, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?16:1);
	
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	glTexCoord2f(-big,-big); glVertex3f(-big,-big,floor);
	glTexCoord2f(+big,-big); glVertex3f(+big,-big,floor);
	glTexCoord2f(+big,+big); glVertex3f(+big,+big,floor);
	glTexCoord2f(-big,+big); glVertex3f(-big,+big,floor);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	
	oglCameraAxes();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Texture Wave Demo"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	camera=vec3(0.5,0.5,0.80);
	camera_orient.y=vec3(0,1,0);camera_orient.z=vec3(0,0,1);
	glutMainLoop(); /* start drawing */
}
